#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
import tempfile
from unittest import TestCase

from liminal.build import liminal_apps_builder
from liminal.kubernetes import volume_util
from liminal.runners.airflow import DummyDag
from liminal.runners.airflow.executors.kubernetes import KubernetesPodExecutor
from liminal.runners.airflow.tasks.spark import SparkTask
from tests.util import dag_test_utils


class TestSparkTask(TestCase):
    """
    Test Spark Task
    """

    _VOLUME_NAME = 'myvol1'

    def test_spark_on_k8s(self):
        volume_util.delete_local_volume(self._VOLUME_NAME)
        os.environ['TMPDIR'] = '/tmp'
        self.temp_dir = tempfile.mkdtemp()
        self.liminal_config = {
            'volumes': [
                {
                    'volume': self._VOLUME_NAME,
                    'local': {
                        'path': self.temp_dir.replace(
                            "/var/folders",
                            "/private/var/folders"
                        )
                    }
                }
            ]
        }
        volume_util.create_local_volumes(self.liminal_config, None)

        # build spark image
        liminal_apps_builder.build_liminal_apps(
            os.path.join(os.path.dirname(__file__), '../../apps/test_spark_app'))

        outputs_dir = os.path.join(self.temp_dir, 'outputs')

        task_config = {
            'task': "my_spark_task",
            'image': "my_spark_image",
            'application_source': 'wordcount.py',
            'application_arguments': ['words.txt', '/mnt/vol1/outputs/'],
            'env_vars': {},
            'mounts': [
                {
                    'mount': 'mymount',
                    'volume': self._VOLUME_NAME,
                    'path': '/mnt/vol1'
                }
            ]
        }

        dag = dag_test_utils.create_dag()

        task1 = SparkTask(
            task_id="my_spark_task",
            dag=dag,
            liminal_config=self.liminal_config,
            pipeline_config={
                'pipeline': 'my_pipeline'
            },
            task_config=task_config,
            parent=None,
            trigger_rule='all_success')

        executor = KubernetesPodExecutor(
            task_id='k8s',
            liminal_config=self.liminal_config,
            executor_config={
                'executor': 'k8s',
                'name': 'mypod'
            }
        )
        executor.apply_task_to_dag(task=task1)

        for task in dag.tasks:
            print(f'Executing task {task.task_id}')
            task.execute(DummyDag('my_dag', task.task_id).context)

        expected_output = '{"word":"my","count":1}\n' \
                          '{"word":"first","count":1}\n' \
                          '{"word":"liminal","count":1}\n' \
                          '{"word":"spark","count":1}\n' \
                          '{"word":"task","count":1}\n'.split("\n")

        actual = ''
        for filename in os.listdir(outputs_dir):
            if filename.endswith(".json"):
                with open(os.path.join(outputs_dir, filename)) as f:
                    actual = f.read()

        self.assertEqual(actual.split("\n"), expected_output)

    def test_get_runnable_command(self):
        task_config = {
            'application_source': 'my_app.py',
            'master': 'yarn',
            'class': 'org.apache.liminal.MySparkApp',
            'conf': {
                'spark.driver.memory': '1g',
                'spark.driver.maxResultSize': '1g',
                'spark.yarn.executor.memoryOverhead': '500M'
            },
            'application_arguments': {
                '--query': "select * from "
                           "my_table where date_prt >= "
                           "'{{yesterday_ds}}'",
                '--output': 'mytable'
            }
        }

        expected = ['spark-submit',
                    '--master',
                    'yarn',
                    '--class',
                    'org.apache.liminal.MySparkApp',
                    '--conf',
                    'spark.driver.maxResultSize=1g', '--conf', 'spark.driver.memory=1g', '--conf',
                    'spark.yarn.executor.memoryOverhead=500M', 'my_app.py',
                    '--query',
                    "select * from my_table where "
                    "date_prt >= '{{yesterday_ds}}'",
                    '--output', 'mytable']

        actual = SparkTask(
            'my_spark_task',
            DummyDag('dag-id', 'my_spark_task'),
            [],
            trigger_rule='all_success',
            liminal_config={},
            pipeline_config={'pipeline': 'pipeline'},
            task_config=task_config
        ).get_runnable_command()

        self.assertEqual(actual.sort(), expected.sort())

    def test_missing_spark_arguments(self):
        task_config = {
            'application_source': 'my_app.py',
            'application_arguments': {
                '--query': "select * from my_table where"
                           " date_prt >= '{{yesterday_ds}}'",
                '--output': 'myoutputtable'
            }
        }

        expected = ['spark-submit', 'my_app.py',
                    '--query',
                    "select * from my_table where "
                    "date_prt >= '{{yesterday_ds}}'",
                    '--output', 'myoutputtable']

        actual = SparkTask(
            'my_spark_task',
            DummyDag('dag-id', 'my_spark_task'),
            [],
            trigger_rule='all_success',
            liminal_config={},
            pipeline_config={'pipeline': 'pipeline'},
            task_config=task_config
        ).get_runnable_command()

        self.assertEqual(actual.sort(), expected.sort())

    def test_partially_missing_spark_arguments(self):
        task_config = {
            'application_source': 'my_app.py',
            'class': 'org.apache.liminal.MySparkApp',
            'conf': {
                'spark.driver.memory': '1g',
                'spark.driver.maxResultSize': '1g',
                'spark.yarn.executor.memoryOverhead': '500M'
            },
            'application_arguments': {
                '--query': "select * from my_table where "
                           "date_prt >= '{{yesterday_ds}}'",
                '--output': 'myoutputtable'
            }
        }

        expected = ['spark-submit',
                    '--class',
                    'org.apache.liminal.MySparkApp',
                    '--conf',
                    'spark.driver.maxResultSize=1g',
                    '--conf',
                    'spark.driver.memory=1g',
                    '--conf',
                    'spark.yarn.executor.memoryOverhead=500M',
                    'my_app.py',
                    '--query',
                    'select * from my_table where '
                    "date_prt >= '{{yesterday_ds}}'",
                    '--output',
                    'myoutputtable']

        actual = SparkTask(
            'my_spark_task',
            DummyDag('dag-id', 'my_spark_task'),
            [],
            trigger_rule='all_success',
            liminal_config={},
            pipeline_config={'pipeline': 'pipeline'},
            task_config=task_config
        ).get_runnable_command()

        self.assertEqual(actual.sort(), expected.sort())



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
import pathlib
from unittest import TestCase

from termcolor import colored

EXCLUDED_EXTENSIONS = ['.gif', '.png', '.pyc', 'LICENSE', 'DISCLAIMER', 'NOTICE', '.whl']
EXCLUDED_DIRS = ['docs/build', '.git', '.idea', 'venv', 'apache_liminal.egg-info']
EXCLUDED_FILES = ['DISCLAIMER-WIP']

PYTHON_LICENSE_HEADER = """
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
""".strip().split("\n")

MD_LICENSE_HEADER = """
<!--
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
-->
""".strip().split("\n")

RST_LICENSE_HEADER = """
..
   Licensed to the Apache Software Foundation (ASF) under one
   or more contributor license agreements.  See the NOTICE file
   distributed with this work for additional information
   regarding copyright ownership.  The ASF licenses this file
   to you under the Apache License, Version 2.0 (the
   "License"); you may not use this file except in compliance
   with the License.  You may obtain a copy of the License at
..

..  http://www.apache.org/licenses/LICENSE-2.0

..
   Unless required by applicable law or agreed to in writing,
   software distributed under the License is distributed on an
   "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
   KIND, either express or implied.  See the License for the
   specific language governing permissions and limitations
   under the License.
..
""".strip().split("\n")

BAT_LICENSE_HEADER = """
REM
REM Licensed to the Apache Software Foundation (ASF) under one
REM or more contributor license agreements.  See the NOTICE file
REM distributed with this work for additional information
REM regarding copyright ownership.  The ASF licenses this file
REM to you under the Apache License, Version 2.0 (the
REM "License"); you may not use this file except in compliance
REM with the License.  You may obtain a copy of the License at
REM
REM   http://www.apache.org/licenses/LICENSE-2.0
REM
REM Unless required by applicable law or agreed to in writing,
REM software distributed under the License is distributed on an
REM "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
REM KIND, either express or implied.  See the License for the
REM specific language governing permissions and limitations
REM under the License.
""".strip().split("\n")

JSON_LICENSE_HEADER = """
  "": [
    "Licensed to the Apache Software Foundation (ASF) under one",
    "or more contributor license agreements.  See the NOTICE file",
    "distributed with this work for additional information",
    "regarding copyright ownership.  The ASF licenses this file",
    "to you under the Apache License, Version 2.0 (the",
    "\\"License\\"); you may not use this file except in compliance",
    "with the License.  You may obtain a copy of the License at",
    "",
    "  http://www.apache.org/licenses/LICENSE-2.0",
    "",
    "Unless required by applicable law or agreed to in writing,",
    "software distributed under the License is distributed on an",
    "\\"AS IS\\"BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY",
    "KIND, either express or implied.  See the License for the",
    "specific language governing permissions and limitations",
    "under the License."
  ],
""".strip().split("\n")


class TestLicenses(TestCase):

    def test_licenses(self):
        files = []
        base_dir = os.path.join(pathlib.Path(__file__).parent.parent.absolute())
        print(f'Checking licenses for files in {base_dir}')
        for r, d, f in os.walk(base_dir):
            if not any(os.path.relpath(r, base_dir).startswith(excluded) for excluded in
                       EXCLUDED_DIRS):
                for file in f:
                    if not any(os.path.basename(file).endswith(ext) for ext in EXCLUDED_EXTENSIONS) \
                            and not os.path.basename(file) in EXCLUDED_FILES:
                        files.append(os.path.join(r, file))

        output = ''
        files_missing_license = []
        success = True
        for file in files:
            print(f'Checking license for file {file}')
            has_license = self.check_license(file)
            if not has_license:
                output += colored(f'Missing License: {file}\n', 'red')
                files_missing_license.append(file)
            success = success and has_license

        print(output)

        if not success:
            self.assertListEqual([], files_missing_license)

    @staticmethod
    def check_license(file):
        header_lines = PYTHON_LICENSE_HEADER
        if file.endswith('.md'):
            header_lines = MD_LICENSE_HEADER
        elif file.endswith('.rst'):
            header_lines = RST_LICENSE_HEADER
        elif file.endswith('.bat'):
            header_lines = BAT_LICENSE_HEADER
        elif file.endswith('.json'):
            header_lines = JSON_LICENSE_HEADER
            header_lines[0] = f'  {header_lines[0]}'
        with open(file) as f:
            file_lines = f.readlines()
        return all(f'{line}\n' in file_lines for line in header_lines)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from datetime import datetime
from unittest import TestCase, mock
from unittest.mock import MagicMock

from airflow.operators.dummy import DummyOperator
from airflow.operators.python import BranchPythonOperator
from moto import mock_cloudformation

from liminal.runners.airflow.executors import airflow
from liminal.runners.airflow.operators.cloudformation import CloudFormationCreateStackOperator, \
    CloudFormationCreateStackSensor, CloudFormationHook
from liminal.runners.airflow.tasks.create_cloudformation_stack import CreateCloudFormationStackTask
from tests.util import dag_test_utils


# noinspection DuplicatedCode
@mock_cloudformation
class TestCreateCloudFormationStackTask(TestCase):
    """
    Test CreateCloudFormationStackTask
    """

    def setUp(self) -> None:
        self.dag = dag_test_utils.create_dag()
        self.dag.context = {
            'ti': self.dag,
            'ts': datetime.now().timestamp(),
            'execution_date': datetime.now().timestamp()
        }
        self.dag.get_dagrun = MagicMock()

        self.cluster_name = "liminal-cluster-for-tests"
        self.config = {
            'task': 'create_emr',
            'type': 'create_cloudformation_stack',
            'description': 'create emr',
            'stack_name': self.cluster_name,
            'properties': {
                'OnFailure': 'DO_NOTHING',
                'TimeoutInMinutes': 25,
                'Capabilities': ['CAPABILITY_NAMED_IAM'],
                'TemplateURL': 'https://s3.amazonaws.com/liminal-tests/emr_cluster_creation.yml',
                'Parameters': {
                    'Environment': 'Staging',
                    'OwnerTeam': 'liminal-team',
                    'Tenancy': 'Ephemeral',
                    'MasterServerCount': '1',
                    'CoreServerCount': '1',
                    'EmrApplicationRelease': '5.28',
                    'InstanceTypeMaster': 'm5.xlarge',
                    'InstanceTypeCore': 'm5.xlarge'
                }
            }
        }

        self.create_cloudformation_task = \
            CreateCloudFormationStackTask(
                self.config['task'],
                self.dag,
                [],
                trigger_rule='all_success',
                liminal_config={},
                pipeline_config={},
                task_config=self.config
            )

        airflow.AirflowExecutor("airflow-executor", {}, {}).apply_task_to_dag(
            task=self.create_cloudformation_task)

    def test_apply_task_to_dag(self):
        self.assertEqual(len(self.dag.tasks), 4)

        self.assertIsInstance(self.dag.tasks[0], BranchPythonOperator)
        self.assertIsInstance(self.dag.tasks[1], CloudFormationCreateStackOperator)
        self.assertIsInstance(self.dag.tasks[2], CloudFormationCreateStackSensor)
        self.assertIsInstance(self.dag.tasks[3], DummyOperator)

    def test_cloudformation_does_not_exist(self):
        with mock.patch.object(CloudFormationHook, 'get_conn') as mock_conn:
            mock_cf_conn = MagicMock()
            mock_cf_conn.describe_stacks.return_value.raiseError.side_effect = Exception()

            mock_conn.return_value = mock_cf_conn
            is_cloudformation_exists = self.dag.tasks[0]

            print(is_cloudformation_exists)
            self.assertEqual(is_cloudformation_exists.python_callable(stack_name=self.cluster_name),
                             'create-cloudformation-create_emr')

    def test_cloudformation_exist_and_running(self):
        is_cloudformation_exists = self.dag.tasks[0]

        for status in ['CREATE_COMPLETE', 'DELETE_FAILED']:
            with mock.patch.object(CloudFormationHook, 'get_conn') as mock_conn:
                mock_cloudformation_conn = MagicMock()
                mock_cloudformation_conn.describe_stacks.return_value = {
                    'Stacks': [
                        {
                            'StackStatus': status
                        }
                    ]
                }

                mock_conn.return_value = mock_cloudformation_conn

                self.assertEqual(
                    is_cloudformation_exists.python_callable(
                        is_cloudformation_exists.python_callable(stack_name=self.cluster_name),
                    'creation-end-create_emr')

    def test_cloudformation_exists_and_not_running(self):
        is_cloudformation_exists = self.dag.tasks[0]

        for status in ['DELETED']:
            with mock.patch.object(CloudFormationHook, 'get_conn') as mock_conn:
                mock_cloudformation_conn = MagicMock()
                mock_cloudformation_conn.describe_stacks.return_value = {
                    'Stacks': [
                        {
                            'StackStatus': status
                        }
                    ]
                }

                mock_conn.return_value = mock_cloudformation_conn

                self.assertEqual(
                    is_cloudformation_exists.python_callable(
                        is_cloudformation_exists.python_callable(stack_name=self.cluster_name),
                    'create-cloudformation-create_emr')

    def test_cloudformation_create_stack_operator_task(self):
        create_cloudformation_stack = self.dag.tasks[1]
        self.assertEqual(create_cloudformation_stack.task_id, 'create-cloudformation-create_emr')



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import logging
import os
import traceback
from datetime import datetime, timedelta

from airflow import DAG

from liminal.core.config.config import ConfigUtil
from liminal.core.util import class_util
from liminal.runners.airflow.executors import airflow
from liminal.runners.airflow.model import executor as liminal_executor
from liminal.runners.airflow.model.task import Task

__DEPENDS_ON_PAST = 'depends_on_past'


# noinspection PyBroadException
def register_dags(configs_path):
    """
    Registers pipelines in liminal yml files found in given path (recursively) as airflow DAGs.
    """
    logging.info(f'Registering DAGs from path: {configs_path}')
    config_util = ConfigUtil(configs_path)
    # TODO - change is_render_variable to False when runtime resolving is available
    configs = config_util.safe_load(is_render_variables=True)

    if os.getenv('POD_NAMESPACE') != "jenkins":
        config_util.snapshot_final_liminal_configs()

    dags = []
    logging.info(f'found {len(configs)} liminal configs in path: {configs_path}')
    for config in configs:
        name = config['name'] if 'name' in config else None
        try:
            if not name:
                raise ValueError('liminal.yml missing field `name`')

            logging.info(f"Registering DAGs for {name}")

            owner = config.get('owner')

            trigger_rule = 'all_success'
            if 'always_run' in config and config['always_run']:
                trigger_rule = 'all_done'

            executors = __initialize_executors(config)

            default_executor = airflow.AirflowExecutor("default_executor", liminal_config=config,
                                                       executor_config={})

            for pipeline in config['pipelines']:
                default_args = __default_args(pipeline)
                dag = __initialize_dag(default_args, pipeline, owner)

                parent = None

                for task in pipeline['tasks']:
                    task_type = task['type']
                    task_instance = get_task_class(task_type)(
                        task_id=task['task'],
                        dag=dag,
                        parent=parent,
                        trigger_rule=trigger_rule,
                        liminal_config=config,
                        pipeline_config=pipeline,
                        task_config=task
                    )

                    executor_id = task.get('executor')
                    if executor_id:
                        executor = executors[executor_id]
                    else:
                        logging.info(f"Did not find `executor` in ${task['task']} config."
                                     f" Using the default executor (${type(default_executor)})"
                                     f" instead.")
                        executor = default_executor

                    parent = executor.apply_task_to_dag(task=task_instance)

                logging.info(f'registered DAG {dag.dag_id}: {dag.tasks}')

                dags.append((pipeline['pipeline'], dag))

        except Exception:
            logging.error(f'Failed to register DAGs for {name}')
            traceback.print_exc()

    return dags


def __initialize_executors(liminal_config):
    executors = {}
    for executor_config in liminal_config.get('executors', {}):
        executors[executor_config['executor']] = get_executor_class(executor_config['type'])(
            executor_config['executor'],
            liminal_config,
            executor_config
        )
    return executors


def __initialize_dag(default_args, pipeline, owner):
    pipeline_name = pipeline['pipeline']

    schedule_interval = default_args.get('schedule_interval', None)
    if not schedule_interval:
        schedule_interval = default_args.get('schedule', None)

    if owner and 'owner' not in default_args:
        default_args['owner'] = owner

    start_date = pipeline.get('start_date', datetime.min.time())
    if not isinstance(start_date, datetime):
        start_date = datetime.combine(start_date, datetime.min.time())

    default_args.pop('tasks', None)
    default_args.pop('schedule', None)
    default_args.pop('monitoring', None)
    default_args.pop('schedule_interval', None)

    dag = DAG(
        dag_id=pipeline_name,
        default_args=default_args,
        dagrun_timeout=timedelta(minutes=pipeline['timeout_minutes']),
        start_date=start_date,
        schedule_interval=schedule_interval,
        catchup=False
    )

    return dag


def __default_args(pipeline):
    default_args = {k: v for k, v in pipeline.items()}
    override_args = {
        'start_date': datetime.combine(pipeline['start_date'], datetime.min.time()),
        __DEPENDS_ON_PAST: default_args[
            __DEPENDS_ON_PAST
        ] if __DEPENDS_ON_PAST in default_args else False,
    }
    default_args.update(override_args)
    return default_args


logging.info(f'Loading task implementations..')

# TODO: add configuration for user tasks package
impl_packages = 'liminal.runners.airflow.tasks'
user_task_package = 'TODO: user_tasks_package'

tasks_by_liminal_name = class_util.find_subclasses_in_packages([impl_packages], Task)

logging.info(f'Finished loading task implementations: {tasks_by_liminal_name.keys()}')

executors_by_liminal_name = class_util.find_subclasses_in_packages(
    ['liminal.runners.airflow.executors'],
    liminal_executor.Executor)

logging.info(f'Finished loading executor implementations: {executors_by_liminal_name.keys()}')


def get_task_class(task_type):
    return tasks_by_liminal_name[task_type]


def get_executor_class(executor_type):
    return executors_by_liminal_name[executor_type]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import sys
from operator import add

from pyspark.sql import SparkSession
import yaml

if __name__ == "__main__":
    if len(sys.argv) != 3:
        print("Usage: wordcount <file> output <outputlocation>", file=sys.stderr)
        sys.exit(-1)

    spark = SparkSession \
        .builder \
        .appName("PythonWordCount") \
        .getOrCreate()

    lines = spark.read.text(sys.argv[1]).rdd.filter(lambda x: not x[0].startswith('#')).map(
        lambda r: r[0])
    counts = lines.flatMap(lambda x: x.split(' ')) \
        .map(lambda x: (x, 1)) \
        .reduceByKey(add)
    output = counts.collect()
    for (word, count) in output:
        print("%s: %i" % (word, count))

    print("writing the results to {}".format(sys.argv[2]))
    counts.toDF(["word", "count"]).coalesce(1).write.mode("overwrite").json(sys.argv[2])

    spark.stop()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from datetime import datetime
from unittest import TestCase
from unittest.mock import MagicMock

from airflow.operators.dummy_operator import DummyOperator
from airflow.operators.python_operator import BranchPythonOperator
from airflow.utils.trigger_rule import TriggerRule
from moto import mock_cloudformation

from liminal.runners.airflow.operators.cloudformation import CloudFormationDeleteStackOperator, \
    CloudFormationDeleteStackSensor
from liminal.runners.airflow.tasks.delete_cloudformation_stack import DeleteCloudFormationStackTask
from tests.util import dag_test_utils


# noinspection DuplicatedCode
@mock_cloudformation
class TestDeleteCloudFormationStackTask(TestCase):
    """
    Test DeleteCloudFormationStackTask
    """

    def setUp(self) -> None:
        self.dag = dag_test_utils.create_dag()
        self.dag.context = {
            'ti': self.dag,
            'ts': datetime.now().timestamp(),
            'execution_date': datetime.now().timestamp()
        }
        self.dag.get_dagrun = MagicMock()

        self.cluster_name = "liminal-cluster-for-tests"

        self.delete_cloudformation_task = \
            DeleteCloudFormationStackTask(
                'delete-emr',
                self.dag,
                [],
                trigger_rule='all_success',
                liminal_config={},
                pipeline_config={},
                task_config={
                    'stack_name': self.cluster_name
                }
            )

        self.delete_cloudformation_task.apply_task_to_dag()

    def test_apply_task_to_dag(self):
        self.assertEqual(len(self.dag.tasks), 4)

        self.assertIsInstance(self.dag.tasks[0], BranchPythonOperator)
        self.assertIsInstance(self.dag.tasks[1], CloudFormationDeleteStackOperator)
        self.assertIsInstance(self.dag.tasks[2], CloudFormationDeleteStackSensor)
        self.assertIsInstance(self.dag.tasks[3], DummyOperator)

    def test_check_dags_queued_task(self):
        self.dag.get_num_active_runs = MagicMock()

        self.dag.get_num_active_runs.return_value = 2
        check_dags_queued_task = self.dag.tasks[0]
        self.assertEqual(check_dags_queued_task.trigger_rule, TriggerRule.ALL_DONE)

        self.assertEqual(check_dags_queued_task.python_callable(),
                         f'delete-end-delete-emr')

        self.dag.get_num_active_runs.return_value = 1

        self.assertEqual(check_dags_queued_task.python_callable(),
                         f'delete-cloudformation-delete-emr')



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
import os

from liminal.build.image_builder import ImageBuilder
from liminal.core.config.config import ConfigUtil
from liminal.core.util import files_util, class_util


def build_liminal_apps(path):
    """
    Build images for liminal apps in path.
    """
    config_util = ConfigUtil(path)
    configs = config_util.safe_load(is_render_variables=True, soft_merge=True)

    for liminal_config in configs:
        base_path = os.path.dirname(files_util.resolve_pipeline_source_file(liminal_config['name']))
        if 'images' in liminal_config:
            for image in liminal_config['images']:
                image_name = image['image']

                if 'source' in image:
                    image_type = image['type']
                    builder_class = __get_image_builder_class(image_type)
                    if builder_class:
                        __build_image(base_path, image, builder_class)
                    else:
                        raise ValueError(f'No such image type: {image_type}')
                else:
                    logging.warning(f'No source configured for image {image_name}.')


def __build_image(base_path, builder_config, builder):
    builder_instance = builder(
        config=builder_config,
        base_path=base_path,
        relative_source_path=builder_config['source'],
        tag=builder_config['image'])
    builder_instance.build()


def __get_image_builder_class(task_type):
    return image_builder_types.get(task_type, None)


logging.info(f'Loading image builder implementations..')

# TODO: add configuration for user image builders package
image_builders_package = 'liminal.build.image'
# user_image_builders_package = 'TODO: user_image_builders_package'

image_builder_types = class_util.find_subclasses_in_packages(
    [image_builders_package],
    ImageBuilder)

logging.info(f'Finished loading image builder implementations: {image_builder_types}')
logging.info(f'Loading service image builder implementations..')



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
from unittest import TestCase, mock

from liminal.core.config.config import ConfigUtil


# noinspection PyUnresolvedReferences,DuplicatedCode
class TestConfigUtil(TestCase):

    @mock.patch('liminal.core.util.files_util.load')
    def test_safe_load(self, find_config_files_mock):
        subliminal = {
            'name': 'my_subliminal_test',
            'type': 'sub',
            'super': 'my_superliminal_test',
            'images': [{
                'image': 'my_image'
            }],
            'pipelines': [
                {'name': 'mypipe1', 'param': 'constant'},
                {'name': 'mypipe2', 'param': 'constant'}
            ],
            'pipeline_defaults': {
                'param1': 'param1_value'
            },
            'task_defaults': {
                'job_start': {
                    'task_sub_def': 'task_sub_def_value'
                }
            }
        }
        superliminal = {
            'name': 'my_superliminal_test',
            'type': 'super',
            'super': 'super_superliminal',
            'images': [{
                'image': 'my_image',
                'source': '.'
            }],
            'pipeline_defaults': {
                'param2': 'param2super_value',
                'param3': 'param3super_value'
            },
            'task_defaults': {
                'job_start': {
                    'task_def1': 'task_def1_value',
                    'task_def2': {
                        'task_def2_1': 'task_def2_1_value',
                    }
                }
            }
        }
        super_superliminal = {
            'name': 'super_superliminal',
            'type': 'super',
            'pipeline_defaults': {
                'param2': 'param2super_value',
                'param3': 'param3hyper_value',
                'param4': 'param4hyper_value'
            }
        }

        expected = [{
            'executors': [{'executor': 'default_k8s', 'type': 'kubernetes'}],
            'name': 'my_subliminal_test',
            'pipeline_defaults': {'param1': 'param1_value'},
            'pipelines': [{'description': 'add defaults parameters for all pipelines',
                           'name': 'mypipe1',
                           'param': 'constant',
                           'param1': 'param1_value',
                           'param2': 'param2super_value',
                           'param3': 'param3super_value',
                           'param4': 'param4hyper_value',
                           'tasks': [{'task': 'start',
                                      'task_def1': 'task_def1_value',
                                      'task_def2': {'task_def2_1': 'task_def2_1_value'},
                                      'task_sub_def': 'task_sub_def_value',
                                      'type': 'job_start'},
                                     {'task': 'end', 'type': 'job_end'}]},
                          {'description': 'add defaults parameters for all pipelines',
                           'name': 'mypipe2',
                           'param': 'constant',
                           'param1': 'param1_value',
                           'param2': 'param2super_value',
                           'param3': 'param3super_value',
                           'param4': 'param4hyper_value',
                           'tasks': [{'task': 'start',
                                      'task_def1': 'task_def1_value',
                                      'task_def2': {'task_def2_1': 'task_def2_1_value'},
                                      'task_sub_def': 'task_sub_def_value',
                                      'type': 'job_start'},
                                     {'task': 'end', 'type': 'job_end'}]}],
            'service_defaults': {'description': 'add defaults parameters for all '
                                                'services'},
            'images': [{'image': 'my_image', 'source': '.'}],
            'services': [],
            'super': 'my_superliminal_test',
            'task_defaults': {'job_start': {'task_sub_def': 'task_sub_def_value'}},
            'type': 'sub'
        }]

        expected = [{'executors': [{'executor': 'default_k8s', 'type': 'kubernetes'},
                                   {'executor': 'airflow_executor', 'type': 'airflow'}],
                     'images': [{'image': 'my_image', 'source': '.'}],
                     'name': 'my_subliminal_test',
                     'pipeline_defaults': {'param1': 'param1_value'},
                     'pipelines': [{'description': 'add defaults parameters for all pipelines',
                                    'name': 'mypipe1',
                                    'param': 'constant',
                                    'param1': 'param1_value',
                                    'param2': 'param2super_value',
                                    'param3': 'param3super_value',
                                    'param4': 'param4hyper_value',
                                    'tasks': [{'executor': 'airflow_executor',
                                               'task': 'start',
                                               'task_def1': 'task_def1_value',
                                               'task_def2': {'task_def2_1': 'task_def2_1_value'},
                                               'task_sub_def': 'task_sub_def_value',
                                               'type': 'job_start'},
                                              {'executor': 'airflow_executor',
                                               'task': 'end',
                                               'type': 'job_end'}]},
                                   {'description': 'add defaults parameters for all pipelines',
                                    'name': 'mypipe2',
                                    'param': 'constant',
                                    'param1': 'param1_value',
                                    'param2': 'param2super_value',
                                    'param3': 'param3super_value',
                                    'param4': 'param4hyper_value',
                                    'tasks': [{'executor': 'airflow_executor',
                                               'task': 'start',
                                               'task_def1': 'task_def1_value',
                                               'task_def2': {'task_def2_1': 'task_def2_1_value'},
                                               'task_sub_def': 'task_sub_def_value',
                                               'type': 'job_start'},
                                              {'executor': 'airflow_executor',
                                               'task': 'end',
                                               'type': 'job_end'}]}],
                     'service_defaults': {'description': 'add defaults parameters for all '
                                                         'services'},
                     'services': [],
                     'super': 'my_superliminal_test',
                     'task_defaults': {'job_start': {'task_sub_def': 'task_sub_def_value'}},
                     'type': 'sub'}]

        find_config_files_mock.return_value = {
            'my_subliminal_test': subliminal,
            'my_superliminal_test': superliminal,
            'super_superliminal': super_superliminal
        }

        config_util = ConfigUtil('')

        self.assertEqual(expected, config_util.safe_load(is_render_variables=True))

        # validate cache
        self.assertEqual(expected, config_util.loaded_subliminals)

    @mock.patch('liminal.core.util.files_util.load')
    def test_get_config(self, find_config_files_mock):
        find_config_files_mock.return_value = {
            'my_subliminal_test': {
                'type': 'sub'
            },
            'my_superliminal_test': {
                'type': 'super'
            }
        }

        config_util = ConfigUtil('')

        self.assertEqual({'type': 'sub'},
                         config_util._ConfigUtil__get_config('my_subliminal_test'))

        self.assertEqual({'type': 'super'},
                         config_util._ConfigUtil__get_config('my_superliminal_test'))

    @mock.patch('liminal.core.util.files_util.load')
    def test_get_superliminal(self, find_config_files_mock):
        base = {'executors': [{'executor': 'default_k8s', 'type': 'kubernetes'},
                              {'executor': 'airflow_executor', 'type': 'airflow'}],
                'name': 'base',
                'pipeline_defaults': {'after_tasks': [{'task': 'end', 'type': 'job_end'}],
                                      'before_tasks': [{'task': 'start', 'type': 'job_start'}],
                                      'description': 'add defaults parameters for all '
                                                     'pipelines'},
                'service_defaults': {'description': 'add defaults parameters for all '
                                                    'services'},
                'task_defaults': {'description': 'add defaults parameters for all tasks '
                                                 'separate by task type',
                                  'job_end': {'executor': 'airflow_executor'},
                                  'job_start': {'executor': 'airflow_executor'},
                                  'python': {'executor': 'default_k8s'},
                                  'spark': {'executor': 'default_k8s'}},
                'type': 'super'}
        subliminal = {
            'name': 'subliminal_test',
            'type': 'sub'
        }

        find_config_files_mock.return_value = {
            'subliminal_test': subliminal
        }

        config_util = ConfigUtil('')

        self.assertEqual(base,
                         config_util._ConfigUtil__get_superliminal(subliminal, False))

        self.assertEqual({},
                         config_util._ConfigUtil__get_superliminal(base, False))

        liminal = {
            'name': 'subliminal_test',
            'type': 'sub',
            'super': 'my_superliminal'
        }

        with self.assertRaises(FileNotFoundError):
            config_util._ConfigUtil__get_superliminal(liminal, False)

    @mock.patch('liminal.core.util.files_util.load')
    def test_merge_superliminals(self, find_config_files_mock):
        superliminal = {
            'name': 'my_superliminal_test',
            'type': 'super',
            'super': 'super_superliminal',
            'pipeline_defaults': {
                'before_tasks': [
                    {'task': 'start-2', 'type': 'spark'},
                ],
                'after_tasks': [
                    {'task': 'end-1', 'type': 'spark'}
                ]
            },
            'task_defaults': {
                'task_def1': 'task_def1_value'
            }
        }

        super_superliminal = {
            'name': 'super_superliminal',
            'type': 'super',
            'pipeline_defaults': {
                'before_tasks': [
                    {'task': 'start-1', 'type': 'spark'}],
                'after_tasks': [
                    {'task': 'end-2', 'type': 'spark'}
                ]
            }
        }

        config_util = ConfigUtil('')

        find_config_files_mock.return_value = {
            'super_superliminal': super_superliminal,
            'superliminal': superliminal
        }

        expected = {
            'name': 'my_superliminal_test',
            'pipeline_defaults': {'after_tasks': [{'task': 'end-1', 'type': 'spark'},
                                                  {'task': 'end-2', 'type': 'spark'}],
                                  'before_tasks': [{'task': 'start-1', 'type': 'spark'},
                                                   {'task': 'start-2', 'type': 'spark'}]},
            'super': 'super_superliminal',
            'task_defaults': {'task_def1': 'task_def1_value'},
            'type': 'super'
        }

        self.assertEqual(expected,
                         dict(config_util._ConfigUtil__merge_superliminals(superliminal,
                                                                           super_superliminal)))

    @mock.patch('liminal.core.util.files_util.load')
    @mock.patch.dict(os.environ, {'env': 'myenv', 'LIMINAL_STAND_ALONE_MODE': 'True'})
    def test_safe_load_with_variables(self, find_config_files_mock):
        subliminal = {
            'name': 'my_subliminal_test',
            'type': 'sub',
            'super': 'my_superliminal_test',
            'variables': {
                'var': 'simple case',
                'var-2': '-case',
                'var_2': '_case',
                'image': 'prod image',
                'a': '{{env}}1',
                'b': '{{a}}2',
                'c': '{{a}}{{b}}2'
            },
            'pipelines': [
                {'name': 'mypipe1', 'param': '{{var}}',
                 'tasks': [
                     {'task': 'sub_tasks',
                      'type': 'dummy'},
                 ]},
                {'name': 'mypipe2', 'param': '{{var-2   }}', 'tasks': [
                    {'task': 'sub_tasks',
                     'type': 'dummy'},
                ]}
            ],
            'pipeline_defaults': {
                'param1': '{{var-2}}'
            },
            'task_defaults': {
                'job_start': {
                    'task_def1:': 'task_sub_def_value'
                }

            },
            'services': [
                {
                    'name': 'my_python_server',
                    'type': 'python_server',
                    'image': '{{image}}'
                },
                {
                    'name': 'my_python_server_for_stg',
                    'type': 'python_server',
                    'image': '{{default_image}}'
                }
            ]}

        superliminal = {
            'name': 'my_superliminal_test',
            'type': 'super',
            'variables': {
                'var-2': 'override',
                'var3': 'super_var',
                'default_image': 'default_image_value',
                'image': 'default_image_value'
            },
            'super': 'super_superliminal',
            'pipeline_defaults': {
                'param2': '{{pipe-var}}',
                'param3': 'param3super_value',
                'before_tasks': [
                    {'task': 'second_task', 'type': 'dummy'},
                ]
            },
            'task_defaults': {
                'pipeline': {
                    'path': '{{var-2}}',
                    'task_def1': 'task_def1_value',
                    'task_def2': {
                        'task_def2_1': 'task_def2_1_value',
                    }
                }
            }
        }
        super_superliminal = {
            'name': 'super_superliminal',
            'type': 'super',
            'variables': {
                'default_image': 'def_default_image_value'
            },
            'pipeline_defaults': {
                'global_conf': '{{var3}}',
                'param2': 'param2super_value',
                'param3': 'param3hyper_value',
                'param4': 'param4hyper_value',
                'after_tasks': [
                    {'task': 'before_last_task', 'type': 'dummy'},
                ]
            }
        }

        expected = [{'executors': [{'executor': 'default_k8s', 'type': 'kubernetes'},
                                   {'executor': 'airflow_executor', 'type': 'airflow'}],
                     'name': 'my_subliminal_test',
                     'pipeline_defaults': {'param1': '-case'},
                     'pipelines': [{'description': 'add defaults parameters for all pipelines',
                                    'global_conf': 'super_var',
                                    'name': 'mypipe1',
                                    'param': 'simple case',
                                    'param1': '-case',
                                    'param2': '{{pipe-var}}',
                                    'param3': 'param3super_value',
                                    'param4': 'param4hyper_value',
                                    'tasks': [{'executor': 'airflow_executor',
                                               'task': 'start',
                                               'task_def1:': 'task_sub_def_value',
                                               'type': 'job_start'},
                                              {'task': 'second_task', 'type': 'dummy'},
                                              {'task': 'sub_tasks', 'type': 'dummy'},
                                              {'task': 'before_last_task', 'type': 'dummy'},
                                              {'executor': 'airflow_executor',
                                               'task': 'end',
                                               'type': 'job_end'}]},
                                   {'description': 'add defaults parameters for all pipelines',
                                    'global_conf': 'super_var',
                                    'name': 'mypipe2',
                                    'param': '-case',
                                    'param1': '-case',
                                    'param2': '{{pipe-var}}',
                                    'param3': 'param3super_value',
                                    'param4': 'param4hyper_value',
                                    'tasks': [{'executor': 'airflow_executor',
                                               'task': 'start',
                                               'task_def1:': 'task_sub_def_value',
                                               'type': 'job_start'},
                                              {'task': 'second_task', 'type': 'dummy'},
                                              {'task': 'sub_tasks', 'type': 'dummy'},
                                              {'task': 'before_last_task', 'type': 'dummy'},
                                              {'executor': 'airflow_executor',
                                               'task': 'end',
                                               'type': 'job_end'}]}],
                     'service_defaults': {'description': 'add defaults parameters for all '
                                                         'services'},
                     'images': [],
                     'services': [{'description': 'add defaults parameters for all services',
                                   'image': 'prod image',
                                   'name': 'my_python_server',
                                   'type': 'python_server'},
                                  {'description': 'add defaults parameters for all services',
                                   'image': 'default_image_value',
                                   'name': 'my_python_server_for_stg',
                                   'type': 'python_server'}],
                     'super': 'my_superliminal_test',
                     'task_defaults': {'job_start': {'task_def1:': 'task_sub_def_value'}},
                     'type': 'sub',
                     'variables': {'a': 'myenv1',
                                   'b': 'myenv12',
                                   'c': 'myenv1myenv122',
                                   'image': 'prod image',
                                   'var': 'simple case',
                                   'var-2': '-case',
                                   'var_2': '_case'}}]

        find_config_files_mock.return_value = {
            'my_subliminal_test': subliminal,
            'my_superliminal_test': superliminal,
            'super_superliminal': super_superliminal
        }

        config_util = ConfigUtil('')

        self.assertEqual(expected, config_util.safe_load(is_render_variables=True))

        # validate cache
        self.assertEqual(expected, config_util.loaded_subliminals)

    @mock.patch('os.path.exists')
    @mock.patch('liminal.core.environment.get_airflow_home_dir')
    @mock.patch('liminal.core.util.files_util.load')
    @mock.patch.dict(os.environ, {'LIMINAL_STAND_ALONE_MODE': 'True', 'POD_NAMESPACE': 'my_pod_ns'})
    def test_liminal_config_snapshot(self, find_config_files_mock,
                                     get_airflow_dir_mock, path_exists_mock):
        subliminal = {
            'name': 'my_subliminal_test',
            'type': 'sub',
            'variables': {
                'var': 1,
                'var-2': True
            },
            'pipelines': [
                {'name': 'mypipe1', 'param': '{{var}}'},
                {'name': 'mypipe2', 'param': '{{var-2   }}'}
            ]
        }

        expected = {'name': 'my_subliminal_test', 'type': 'sub',
                    'executors': [{'executor': 'default_k8s', 'type': 'kubernetes'},
                                  {'executor': 'airflow_executor', 'type': 'airflow'}],
                    'service_defaults': {'description': 'add defaults parameters for all services'},
                    'task_defaults': {
                        'description': 'add defaults parameters for all tasks separate by task type',
                        'python': {'executor': 'default_k8s'}, 'spark': {'executor': 'default_k8s'},
                        'job_end': {'executor': 'airflow_executor'},
                        'job_start': {'executor': 'airflow_executor'}}, 'pipeline_defaults': {
                'description': 'add defaults parameters for all pipelines',
                'before_tasks': [{'task': 'start', 'type': 'job_start'}],
                'after_tasks': [{'task': 'end', 'type': 'job_end'}]},
                    'variables': {'var': 1, 'var-2': True}, 'pipelines': [
                {'name': 'mypipe1', 'param': '1',
                 'description': 'add defaults parameters for all pipelines',
                 'tasks': [{'task': 'start', 'type': 'job_start', 'executor': 'airflow_executor'},
                           {'task': 'end', 'type': 'job_end', 'executor': 'airflow_executor'}]},
                {'name': 'mypipe2', 'param': 'True',
                 'description': 'add defaults parameters for all pipelines',
                 'tasks': [{'task': 'start', 'type': 'job_start', 'executor': 'airflow_executor'},
                           {'task': 'end', 'type': 'job_end', 'executor': 'airflow_executor'}]}],
                    'images': [],
                    'services': []}

        find_config_files_mock.return_value = {
            'my_subliminal_test': subliminal
        }

        get_airflow_dir_mock.return_value = '/tmp'
        path_exists_mock.return_value = True

        with mock.patch('builtins.open', mock.mock_open()) as m:
            with mock.patch('yaml.dump') as ydm:
                config_util = ConfigUtil('')
                config_util.safe_load(is_render_variables=True)
                config_util.snapshot_final_liminal_configs()

                m.assert_called_once_with(
                    os.path.join('/tmp', '../liminal_config_files/my_subliminal_test.yml'), 'w')
                ydm.assert_called_once_with(expected, m.return_value, default_flow_style=False)

    @mock.patch('liminal.core.util.files_util.load')
    def test_soft_merge_load(self, find_config_files_mock):
        subliminal = {
            'name': 'my_name',
            'type': 'sub',
            'super': 'my_super'
        }
        find_config_files_mock.return_value = {'my_subliminal_test': subliminal}

        config_util = ConfigUtil('')

        self.assertEqual([subliminal],
                         config_util.safe_load(is_render_variables=True, soft_merge=True))

    def test_non_soft_merge_load(self):
        subliminal = {
            'name': 'my_name',
            'type': 'sub',
            'super': 'my_super'
        }

        config_util = ConfigUtil('')

        self.assertRaises(FileNotFoundError,
                          config_util._ConfigUtil__get_superliminal,
                          subliminal,
                          False)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from airflow.operators.dummy_operator import DummyOperator
from airflow.operators.python_operator import BranchPythonOperator
from airflow.utils.trigger_rule import TriggerRule

from liminal.runners.airflow.operators.cloudformation import CloudFormationDeleteStackOperator, \
    CloudFormationDeleteStackSensor
from liminal.runners.airflow.tasks import airflow


class DeleteCloudFormationStackTask(airflow.AirflowTask):
    """
    Deletes cloud_formation stack.
    """

    def __init__(self, task_id, dag, parent, trigger_rule, liminal_config, pipeline_config,
                 task_config):
        super().__init__(task_id, dag, parent, trigger_rule, liminal_config,
                         pipeline_config, task_config)
        self.stack_name = task_config['stack_name']

    def apply_task_to_dag(self):
        check_dags_queued_task = BranchPythonOperator(
            task_id=f'{self.task_id}-is-dag-queue-empty',
            python_callable=self.__queued_dag_runs_exists,
            provide_context=True,
            trigger_rule=TriggerRule.ALL_DONE,
            dag=self.dag
        )

        delete_stack_task = CloudFormationDeleteStackOperator(
            task_id=f'delete-cloudformation-{self.task_id}',
            params={'StackName': self.stack_name},
            dag=self.dag
        )

        delete_stack_sensor = CloudFormationDeleteStackSensor(
            task_id=f'cloudformation-watch-{self.task_id}-delete',
            stack_name=self.stack_name,
            dag=self.dag
        )

        stack_delete_end_task = DummyOperator(
            task_id=f'delete-end-{self.task_id}',
            dag=self.dag
        )

        if self.parent:
            self.parent.set_downstream(check_dags_queued_task)

        check_dags_queued_task.set_downstream(stack_delete_end_task)
        check_dags_queued_task.set_downstream(delete_stack_task)
        delete_stack_task.set_downstream(delete_stack_sensor)
        delete_stack_sensor.set_downstream(stack_delete_end_task)

        return stack_delete_end_task

    # noinspection PyUnusedLocal
    def __queued_dag_runs_exists(self, **kwargs):
        if self.dag.get_num_active_runs() > 1:
            return f'delete-end-{self.task_id}'
        else:
            return f'delete-cloudformation-{self.task_id}'



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from airflow.operators.dummy import DummyOperator
from airflow.operators.python import BranchPythonOperator
from flatdict import FlatDict

from liminal.runners.airflow.operators.cloudformation import CloudFormationCreateStackOperator, \
    CloudFormationCreateStackSensor, CloudFormationHook
from liminal.runners.airflow.tasks import airflow


class CreateCloudFormationStackTask(airflow.AirflowTask):
    """
    Creates cloud_formation stack.
    """

    def __init__(self, task_id, dag, parent, trigger_rule, liminal_config, pipeline_config,
                 task_config):
        super().__init__(task_id, dag, parent, trigger_rule, liminal_config,
                         pipeline_config, task_config)
        self.stack_name = task_config['stack_name']

    def apply_task_to_dag(self):
        check_cloudformation_stack_exists_task = BranchPythonOperator(
            op_kwargs={'stack_name': self.stack_name},
            task_id=f'is-cloudformation-{self.task_id}-running',
            python_callable=self.__cloudformation_stack_running_branch,
            dag=self.dag
        )

        create_cloudformation_stack_task = CloudFormationCreateStackOperator(
            task_id=f'create-cloudformation-{self.task_id}',
            params={
                **self.__reformatted_params()
            },
            dag=self.dag
        )

        create_stack_sensor_task = CloudFormationCreateStackSensor(
            task_id=f'cloudformation-watch-{self.task_id}-create',
            stack_name=self.stack_name,
            dag=self.dag
        )

        stack_creation_end_task = DummyOperator(
            task_id=f'creation-end-{self.task_id}',
            dag=self.dag,
            trigger_rule='all_done'
        )

        if self.parent:
            self.parent.set_downstream(check_cloudformation_stack_exists_task)

        create_stack_sensor_task.set_downstream(stack_creation_end_task)
        create_cloudformation_stack_task.set_downstream(create_stack_sensor_task)
        check_cloudformation_stack_exists_task.set_downstream(create_cloudformation_stack_task)
        check_cloudformation_stack_exists_task.set_downstream(stack_creation_end_task)

        return stack_creation_end_task

    def __cloudformation_stack_running_branch(self, stack_name):
        cloudformation = CloudFormationHook().get_conn()
        try:
            stack_status = cloudformation.describe_stacks(StackName=stack_name)['Stacks'][0][
                'StackStatus']
            if stack_status in ['CREATE_COMPLETE', 'DELETE_FAILED']:
                print(f'Stack {stack_name} is running')
                return f'creation-end-{self.task_id}'
            else:
                print(f'Stack {stack_name} is not running')
        except Exception as e:
            if 'does not exist' in str(e):
                print(f'Stack {stack_name} does not exist')
                return f'create-cloudformation-{self.task_id}'
            else:
                raise e

        return f'create-cloudformation-{self.task_id}'

    # noinspection PyUnusedLocal
    def __reformatted_params(self, **kwargs):
        return {
            'StackName': self.stack_name,
            **self.task_config['properties'],
            'Parameters': [{'ParameterKey': x, 'ParameterValue': str(y)} for (x, y) in
                           FlatDict(self.task_config['properties']['Parameters']).items()]
        }
